// Copyright (c) 2017, Tobias Mueller tm(at)tm3d.de
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the
//    distribution.
//  * All advertising materials mentioning features or use of this
//    software must display the following acknowledgement: This product
//    includes software developed by tm3d.de and its contributors.
//  * Neither the name of tm3d.de nor the names of its contributors may
//    be used to endorse or promote products derived from this software
//    without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <mysql.h>
#include "mySensorDB.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/mman.h>
#include <fcntl.h>    /* For O_RDWR */
#include <unistd.h>
#include <pthread.h>
#include <sched.h>
#include <memory.h>
#include "owTools.h"
#include "hexfile.h"
#include <iostream>
#include <algorithm>
#include <fstream>
#include <sstream> 



#define RB "\e(0\x6a\e(B" // 188 Right Bottom corner
#define RT "\e(0\x6b\e(B" // 187 Right Top corner
#define LT "\e(0\x6c\e(B" // 201 Left Top cornet
#define LB "\e(0\x6d\e(B" // 200 Left Bottom corner
#define MC "\e(0\x6e\e(B" // 206 Midle Cross
#define HL "\e(0\x71\e(B" // 205 Horizontal Line
#define LC "\e(0\x74\e(B" // 204 Left Cross
#define RC "\e(0\x75\e(B" // 185 Right Cross
#define BC "\e(0\x76\e(B" // 202 Bottom Cross
#define TC "\e(0\x77\e(B" // 203 Top Cross
#define VL "\e(0\x78\e(B" // 186 Vertical Line
#define SP " " 		  // space string
        

void printhelp() {
printf("owTools - Programm for reading und  controlling 1-Wire Devices from www.tm3d.de\n\n");
printf("run:  owTools -a [COMn|USBn|GPIOn] [options]\n\n");
printf("    COMn -> Adapter DS9097 and compatible (e.g. LinkUSB)\n");
printf("         n=1 -> Windows COM1 -> Linux /dev/ttyS1\n");
printf("    USBn -> Adapter DS2490/DS9490 \n");
printf("         n=1 -> first USB-Adapter \n");
printf("    GPIOn -> port of Raspberry PI (port Name not Pin number)\n\n");
printf("options:\n");
printf("   -i interactive mode\n");
printf("          select a device and get information about it\n");
printf("   -c read and print all continuous\n");
printf("      -p time in sec between readings\n");
printf("      -d [config file] put all Data in mysql Database descripted in Config file\n");
printf("         Config file: \n");
printf("            [server]\n");
printf("            [port] -> 3306\n");
printf("            [user] \n");
printf("            [password]\n");
printf("            [database]\n");
printf("            [prefix] -> a prefix of all tables \n\n");
printf("      -r Search everytime for new Devices\n");
printf("   -f hexfile  Flash new\n");
printf("   -n change id   \n");
printf("      -g get from server\n");
printf("   -w [1|2] show Warnings (1) or all Infos (2)\n");



}


typedef struct {
	std::string flag;
	std::string arg;
} arg_t ;


//length of shown chars of an UTF8 string ... sometimes not the same as byte count
std::size_t utf8_length(std::string const &s) {
  std::size_t len = 0;
  std::string::const_iterator begin = s.begin(), end = s.end();
  while (begin != end) {
    unsigned char c = *begin;
    int n=1;
    if      ((c & 0x80) == 0)    n = 1;
    else if ((c & 0xE0) == 0xC0) n = 2;
    else if ((c & 0xF0) == 0xE0) n = 3;
    else if ((c & 0xF8) == 0xF0) n = 4;
    
    len += 1;
    begin += n;
  }
  return len;
}


std::vector<arg_t> comlist;

std::string getArg(std::string flag) {
	for(arg_t a:comlist) {
		if (a.flag==flag) return a.arg;
	}
	return ""; 
}



int getArgi(std::string flag) {
	return atoi(getArg(flag).c_str());
}


int database=0;
mySensorDB *sdb=NULL;


snum_t getArgsnum(std::string flag) {
	snum_t snum;
	snum.num=0;
	return snum;
}

int findCPU(std::string cpu) {
	std::ifstream fileInput;
	std::string line;
	fileInput.open("/proc/cpuinfo");
	if(fileInput.is_open()) {	
		for(unsigned int curLine = 0; getline(fileInput, line); curLine++) {
			if (line.find(cpu) != std::string::npos) {
				fileInput.close();
				return 1;
			}
		}
		fileInput.close();
		
	}
	return 0;
}




owInterface *owi= NULL;

void setLogMode() {
	if (getArg("w")=="2")
			owi->log->setLogLevel(0);
	else if (getArg("w")=="2") owi->log->setLogLevel(0);
	else owi->log->setLogLevel(OWLOG_ERROR);

}


void continuous(std::vector<owDevice*> *devices,int intervall,int headline,int searchrom) {
	int first=1;
	while (1) {
		if (searchrom) {
			owi->log->clear();
			owi->Find();
			if (owi->isDevicesChanged()||(first==1)) {
				first=0;
				if (headline) {
					printf("                     \t");
					for (owDevice* dev :*devices) {
						if (dev->configstate!=OWDCS_NONE) {
							for (size_t i=0;i<4;i++) {
								if (dev->config->getPropertyID(i)!=0)
									printf("\033[1;34m%02X.%02X%02X\033[0m\t",dev->getNum().byte[7],dev->getNum().byte[1],dev->getNum().byte[0]);
							}
						}
					}
					printf("\ntime                 \t");
					for (owDevice* dev :*devices) {
						if (dev->configstate!=OWDCS_NONE) {
							for (size_t i=0;i<4;i++) {
								if (dev->config->getPropertyID(i)!=0)
									printf("\033[0;36m%s\033[0m\t",dev->config->getQuantity(i).substr(0,7).c_str());
							}
						}
					}
					printf("\n                     \t");
					for (owDevice* dev :*devices) {
						if (dev->configstate!=OWDCS_NONE) {
							for (size_t i=0;i<4;i++) {
								if (dev->config->getPropertyID(i)!=0)
									printf("\033[3;34m[%s]  \033[0m\t",dev->config->getUnit(i).c_str());
							}
						}
					}
					
				}	
				printf("\n");

				if (database) {
					//owi->log->setLogLevel(0);
					for (owDevice* dev : *devices) {
						sdb->createDeviceTable(dev);
					}
				}
			}
			
		}
		time_t t=time(NULL);
		int st=(int)t;
		struct tm tm = *localtime(&t);
		printf("%d-%02d-%02d %02d:%02d:%02d\t", tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);
		for (owDevice* dev :*devices) {
			for(int k=0;k<owi->maxrepeat;k++){
				dev->convertAll();
				if (owi->log->last()<OWLOG_ERROR) break;
				owi->log->clear();
			}
			if (owi->log->last()<OWLOG_ERROR) {
				if (dev->configstate!=OWDCS_NONE) {
					for (size_t i=0;i<4;i++) {					
						if (dev->config->getPropertyID(i)!=0) {
							if (dev->values[i]<10) 
								printf("%0.4f \t",dev->values[i]);
							else if (dev->values[i]<100)	
								printf("%0.3f \t",dev->values[i]);
							else if (dev->values[i]<1000)	
								printf("%0.2f \t",dev->values[i]);
							else
								printf("%0.1f \t",dev->values[i]);
							fflush(stdout);
						}
					}
				}
				if (database) {
					sdb->insertValues(dev);
				}
			} else {
				owi->log->setLogLevel(OWLOG_INFO);
				owi->log->set(OWLOG_ERROR,"Too many errors, mybee conection is dead.");
				return;
			}
			
		
		}
		printf("\n");
		while (((int)time(NULL))<(st+intervall)) sleep(1);
	}
}

void device_menu(owDevice* d) {
	for(int i=0;i<70;i++) printf(HL);printf("\n");
	printf("Selected Device: ");
	snum_t snum=d->getNum();
	printf("\033[1;34m%016llX\033[0m",(unsigned long long)snum.num);
	printf("  \n");
	printf("Chip Info: ");
	std::vector<std::string> info=d->getFamilyInfo();
	for (std::string s : info) printf(" %s",s.c_str());
	printf("\n");
	printf("Values info: ");
	int tm3d=d->readConfig();
	for (size_t i=0;i<4;i++) {
		printf("%s in %s",d->config->getQuantity(i).c_str(),d->config->getUnit(i).c_str());
		if (i<4) printf("; "); 
	}
	if (tm3d) printf(" (tm3d.de)"); else printf(" (default)"); 
	printf("\n");
	d->convertAll();
	for (size_t i=0;i<d->values.size();i++) {
		printf("\033[1;33m%0.4f %s\033[0m   ",d->values[i],d->config->getUnit(i).c_str());
	}
	printf("\n");
	for(int i=0;i<70;i++) printf(HL);printf("\n");
	printf("0) Exit\n");
	printf("1) Run continuous\n");
	std::string inp;
	while (1) {
		printf("Select an option: ");std::getline (std::cin,inp);
		int dnr=atoi(inp.c_str());
		if (dnr<=0) return;
		if (dnr==1) {
			std::vector<owDevice*> v;
			v.push_back(d);
			continuous(&v,1,1,0);
		} 
//			if (dnr<=i) break; 
	//		printf("\033[4;33mSelect a number between 0 and %i\033[0m\n",i);
	}
	
	
}

int selectDevice() {
	owi->Find();
	int i=0;
	for (owDevice* dev :owi->devices) {
		i++;
		snum_t snum=dev->getNum();
		printf("%i) ",i);
		printf("\033[1;34m%016llX\033[0m",(unsigned long long)snum.num);

		printf("  ");
		for (int j=0;j<4;j++) printf("%s ",dev->config->getQuantity(j).c_str());
		printf("\n");
	}
	printf("0) Exit Programm\n");
	std::string inp;
	int dnr=0;
	while (1) {
		printf("Select device with number: ");std::getline (std::cin,inp);
		dnr=atoi(inp.c_str());
		if (dnr<=0) return 0;
		if (dnr<=i) break; 
		printf("\033[4;33mSelect a number between 0 and %i\033[0m\n",i);
	}
	return dnr;

}

int questionYesNo(std::string text) {
	std::string inp;
	printf("%s [Y/n] ",text.c_str());std::getline (std::cin,inp);
	if (inp=="Y") return 1; else return 0;
	
}

void interactive(owInterface *owi) {
	while (1) {
		int i=selectDevice();
		if (i==0 ) return;
		device_menu(owi->devices[i-1]);
	}
}

int main(int argc, char *argv[]) {
	int i;
	arg_t a;
	//printf("Kommando: %s:\n", argv[0]);
	while (--argc) {
		if (**++argv != '-') {
      	//printf("   Arg. = %s\n", *argv);
      	if (comlist.empty()) {
      		a.flag=".";
      		a.arg=*argv;
      		comlist.push_back(a);
       	} else {
       		if ((*(comlist.end()-1)).arg!="1") {
      			a.flag=".";
   	   		a.arg=*argv;
	      		comlist.push_back(a);
       		} else {
       			(*(comlist.end()-1)).arg=*argv;
       		}
       	}
		} else {
			for( i=1; (*argv)[i] != '\0'; i++) {
				//printf("   Flag = (%c)\n", (*argv)[i]);
				a.flag=(*argv)[i];
				a.arg="1";
				comlist.push_back(a);
			}
		}
    }          

	//for(arg_t a:comlist) printf("%s->%s\n",a.flag.c_str(),a.arg.c_str());
	if (getArg("h")=="1") {
		printhelp();
		return 0;
	}
	std::string adapter=getArg("a");
	if (adapter.empty()) {
		printhelp();
		return 0;
	}	
	std::transform(adapter.begin(), adapter.end(),adapter.begin(), ::toupper);

	std::string s;
	
	if(adapter.find("COM")!=std::string::npos) {
		owi=new owCOMInterface();
		int port=atoi(adapter.substr(adapter.find("COM")+3).c_str());
		printf("Open /dev/ttyS%i\n",port);
		owi->InitAdapter(port);
	} else 
	if(adapter.find("USB")!=std::string::npos) {
		owi=new owUSBInterface();
		int port=atoi(adapter.substr(adapter.find("USB")+3).c_str());
		printf("Open the %i. USB Adapter\n",port);
		int err;
		if ((err=owi->InitAdapter(port))<0) {
			if (err==-1) printf("No Adapter found!\n  On Linux, try: sudo owTools....\n  On Windows: Install libusb0 driver\n");
			if (err==-2) printf("Maybe Adapter is used ... (try sudo)\nOn Linux: If the kernel modul is loaded type: sudo rmmod ds2490\n");
			exit(0);
		}
	} else 
	if(adapter.find("GPIO")!=std::string::npos) {
		if (findCPU("BCM2708")) {
			printf("\033[1;33mRaspberry Pi 1 is dedected. 1-wire on the GPIO port can have many errors in the transfer.\033[0m\n");
			owi=new owPiGPioInterface();
			int port=atoi(adapter.substr(adapter.find("GPIO")+4).c_str());
			printf("Open GPIO %i\n",port);
		} else if (findCPU("BCM2709")) {
			owi=new owPiGPioInterface();
			int port=atoi(adapter.substr(adapter.find("GPIO")+4).c_str());
			printf("Open GPIO %i\n",port);
			owi->InitAdapter(port);
		} else {
			printf("\033[1;31mGPIO works with Raspberry PI only \033[0m\n");
		}
	}  
	if (owi==NULL) {
		return 0;
	}

	setLogMode();


	if (getArg("i")=="1") {
		interactive(owi);	
	} else 
	if (getArg("c")=="1") {
		int reload=(getArg("r")=="1");
		owi->Find();
		printf("\n");
		//owi->log->setLogLevel(3);
		std::string s;
		int pause;
		if ((s=getArg("p")) !="") {
			pause=atoi(s.c_str());
		} else pause=30;
		if  ((s=getArg("d")) !="") {
			reload=1;
			printf("Use Database\n");
			sdb=new mySensorDB(s,owi->log);
			if (sdb->log->last()>OWLOG_WARNING) {
				exit(1);
			}
			sdb->connect();
			if (sdb->log->last()>OWLOG_WARNING) {
				exit(1);
			}
			sdb->createSensorTable();
			if (sdb->log->last()>OWLOG_WARNING) {
				exit(1);
			}				
			database=1;
		}	
			
		
		continuous(&(owi->devices),pause,1,reload);	
	} else 
	if ((s=getArg("f"))!="") {
		printf("Flash %s in selected Device\n",s.c_str());
		int sel=selectDevice();
		if (sel==0) exit(0);
		int resetid=questionYesNo("Would you reset to default ID? (This is necessary if another device is used!)");
		owi->log->setLogLevel(OWLOG_INFO);		
		owi->flashHEXFile(s,owi->device_nums[sel-1],resetid,1);
		setLogMode();
		owi->Clean();
		owi->Find();
		for (owDevice* dev :owi->devices) {
			snum_t snum=dev->getNum();
			printf("\033[1;34m%016llX\033[0m",(unsigned long long)snum.num);

			printf("\n");
		}
		printf("%i Devices found\n",(int)owi->devices.size());	
				
	} else 
	
	if ((s=getArg("n"))!="") {
		printf("Change id of selected Device\n");
		int sel=selectDevice();
		if (sel==0) exit(0);
		sel-=1;
		
		snum_t snum=owi->devices[sel]->getNum();
		if ((getArg("g"))=="1") {
			printf("get ID from Server\n");
			char s[255];
			sprintf(s,"wget -q http://www.tm3d.de/shop99/owid.php?fam=%02X -O id.txt",snum.byte[0]);
			int err;
			if ((err=system(s))==0) {
				printf("OK!\n");
				std::string rs;
				std::stringstream str;
				std::ifstream fs;
				fs.open ("id.txt" );
				snum_t isnum;
				int i=0; int br=0;
				if(fs.is_open()) {
	    			while(fs.peek() != EOF) {
	    				char c= (char) fs.get();
	    				if (br==0) {
	    					if (c=='x') br=1;
	    				} else {
	    					if ((c==',')|(c=='}')) {
	    						isnum.byte[i]=strtol(s, NULL, 16);
	    						//printf("%x\n",strtol(s, NULL, 16));
	    						i++;
	    						br=0;
	    						s[0]=0;
	    					} else {
	    						s[br-1]=c;
	    						s[br]=0;
	    						br++;
	    					}
	    				}
	    				if (c=='}') break;
	    			}
	    			
	    			fs.close();
				}			 		  
				snum.num=isnum.num;
			} else (printf("ERROR %i\n",err));
		} else {
			snum.num+=256;
		}


		printf("New ID: ");
		printf("\033[1;34m%016llX\033[0m",(unsigned long long)snum.num);

		printf("\n");
		
		owi->devices[sel]->changeID(snum);		
	
		owi->Clean();
		owi->Find();
		for (owDevice* dev :owi->devices) {
			snum_t snum=dev->getNum();
			printf("\033[1;34m%016llX\033[0m",(unsigned long long)snum.num);

			printf("\n");
		}
		printf("%i Devices found\n",(int)owi->devices.size());	
				
	} else {
		owi->Find();
		for (owDevice* dev :owi->devices) {
			printf("_______________________________________________________________________\n");
			snum_t snum=dev->getNum();
			printf("\033[1;34m%016llX\033[0m ",(unsigned long long)snum.num);
			switch (dev->configstate) {
				case OWDCS_16:printf("old Configcode from www.tm3d.de");break;
				case OWDCS_24:printf("Configcode from www.tm3d.de");break;
				case OWDCS_DEFAULT:printf("Default Configcode set by this Software");break;
				case OWDCS_NONE:printf("No Configcode");break;
			}
			printf( " - (");
			for(int i=0;i<3;i++)
				printf(" %s",dev->getFamilyInfo()[i].c_str());
			printf(" )\n");
			if (dev->configstate!=OWDCS_NONE) {
				std::vector<std::vector<std::string>> ia;
				std::vector<size_t> maxcol;
				dev->convertAll();
				for(int i=0;i<4;i++) {
					std::vector<std::string> colm;
					size_t max=0;
					colm.push_back(dev->config->getSensorChip(i));
					if (max<utf8_length(colm[0])) max=utf8_length(colm[0]);
					colm.push_back(dev->config->getQuantity(i));
					if (max<utf8_length(colm[1])) max=utf8_length(colm[1]);
					colm.push_back(dev->config->getUnit(i));
					if (max<utf8_length(colm[2])) max=utf8_length(colm[2]);
					if (dev->config->getPropertyID(i)==0) {
						colm.push_back("");
					} else {
						char hs[50];
						sprintf(hs,"%0.3f",dev->values[i]);
						colm.push_back(hs);
					}
					if (max<utf8_length(colm[3])) max=utf8_length(colm[3]);
					ia.push_back(colm);
					maxcol.push_back(max);
				}
				std::string space="                                                          ";
				for(int i=0;i<4;i++) {
					for(int j=0;j<4;j++) {
						printf("%s%s",ia[j][i].c_str(),space.substr(0,maxcol[j]-utf8_length(ia[j][i])+2).c_str());
					}
					printf("\n");
				}
			}
		}
		printf("%i Devices found\n",(int)owi->devices.size());
	}
	return 0;
   
}
