// Copyright (c) 2017, Tobias Mueller tm(at)tm3d.de
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the
//    distribution.
//  * All advertising materials mentioning features or use of this
//    software must display the following acknowledgement: This product
//    includes software developed by tm3d.de and its contributors.
//  * Neither the name of tm3d.de nor the names of its contributors may
//    be used to endorse or promote products derived from this software
//    without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



#ifndef __OWDEVICECONFIG_H_
#define __OWDEVICECONFIG_H_

#include "owInterface.h"

struct owd_val {
	uint8_t prop;
	uint8_t form;
};

struct owd_config {
	owd_val val[4];
	uint8_t uc;
	uint8_t sens[4];
	uint8_t sp[4];
	uint8_t id2[7];
	uint8_t crc;
	uint8_t crc2;
};
	

#define OWDC_LANGUAGE_ENGLISH 0
#define OWDC_LANGUAGE_GERMAN 1

class owDeviceConfig { 
	owd_config config;
	int language;
public:
	int valid;
	owDeviceConfig() {
		language=OWDC_LANGUAGE_ENGLISH;
		setInvalid();
	};
	void checkValid() {
		for(int i=0;i<4;i++) {
			if ((config.val[i].prop==0xFF)||(config.val[i].form==0xFF)) {valid=0;return;}
			if (config.val[i].prop!=0) valid=1;			
		}
	}
	void setInvalid() {
		valid=0;	
	}
	void setLanguage(int lang) {
		language=lang;
	}
	void setConfig(std::vector<uint8_t> v) {
		int i=0;
		for (uint8_t w :v) {
			((uint8_t*)&config)[i]=w;
			i++;
			if (i>=26) break;
		}
		checkValid();
	}
	void setConfig(const uint8_t v[24]) {
		for (int i=0;i<24;i++) {
			((uint8_t*)&config)[i]=v[i];
			if (i>=26) break;
		}
		checkValid();
	}
	static std::vector<std::string> getFamilyInfo(uint8_t famcode);
	std::vector<std::string> getQuantityFromNumber(uint8_t code);
	double calculateValueFromNumber(int code, int vn, std::vector<int> V);

	std::string getQuantity(uint8_t ch) {
		if (ch>4) return getQuantityFromNumber(255)[0];
		return getQuantityFromNumber(config.val[ch].prop)[0];
	}
	std::string getUnit(uint8_t ch) {
		if (ch>4) return getQuantityFromNumber(255)[0];
		return getQuantityFromNumber(config.val[ch].prop)[1];
	}
	std::string getSensorChip(uint8_t ch);
	double calculateValue(int ch, std::vector<int32_t> V) {
		if (ch>4) return 0;
		return calculateValueFromNumber(config.val[ch].form,ch,V);
	}
	uint8_t getPropertyID(uint8_t ch) {
		if (ch>4) return 0;
		return config.val[ch].prop;
	}
	
	uint8_t getFormulaID(uint8_t ch) {
		if (ch>4) return 0;
		return config.val[ch].form;
	}
	
	
};
#endif 

